@@ -21,6 +21,24 @@ if TYPE_CHECKING:
 def _identity(x): return x
 
 class Study(UserDict[str, sitk.Image | Any]):
+    """A dictionary of scans, segmentations and other info.
+
+    Segmentations should have keys starting with ``"seg"``, for example ``"seg_tumor"``.
+    Segmentation should NOT be one-hot encoded.
+
+    Non-image values should have keys starting with ``"info"``, for example ``"info_id=12345"``.
+
+    You can pass:
+    - path to file to be opened with SimpleITK (e.g. nii/nii.gz);
+    - path to a DICOM dir, make sure it contains one series.
+    - ``sitk.Image``
+    - ``np.ndarray``
+    - ``torch.Tensor``.
+
+    Everything except info will be converted to ``sitk.Image``.
+
+    Please note that multi-channel scans are currently not supported.
+    """
     @overload
     def __init__(self, /, **kwargs): ...
     @overload
@@ -61,6 +79,19 @@ class Study(UserDict[str, sitk.Image | Any]):
         study[key] = item
         return study
 
+    def remove(self, *keys: str | Sequence[str]):
+        """Returns a new study without specified keys"""
+        keys_proc = []
+        for k in keys:
+            if isinstance(k, str): keys_proc.append(k)
+            else: keys_proc.extend(k)
+
+        study = self.copy()
+        for k in keys_proc:
+            del study[k]
+
+        return study
+
     def get_scans(self):
         """Returns a new ``Study`` with segmentations and info removed."""
         return self.__class__({k:v for k,v in self.items() if not k.startswith(("seg", "info"))})
@@ -119,7 +150,7 @@ class Study(UserDict[str, sitk.Image | Any]):
         return self.cast(sitk.sitkFloat32)
 
     def normalize(self) -> "Study":
-        """Return a new study where all scans are separately z-normalized to 0 mean and 1 variance.
+        """Returns a new study where all scans are separately z-normalized to 0 mean and 1 variance.
 
         Note:
             This operation does not affect segmentations.
@@ -127,7 +158,7 @@ class Study(UserDict[str, sitk.Image | Any]):
         return self.apply(sitk.Normalize, seg_fn=None)
 
     def rescale_intensity(self, min: float, max: float) -> "Study":
-        """Return a new study where all scans are separately rescaled to the specified intensity range.
+        """Returns a new study where all scans are separately rescaled to the specified intensity range.
 
         Args:
             min: Minimum value for the output intensity range.
@@ -136,76 +167,150 @@ class Study(UserDict[str, sitk.Image | Any]):
         return self.apply(partial(sitk.RescaleIntensity, outputMinimum = min, outputMaximum = max), seg_fn=None) # type:ignore
 
     def crop_bg(self, key: str) -> "Study":
-        """Return a new study with cropped black background. Finds the foreground bounding box of ``study[key]``,
+        """Returns a new study with cropped black background. Finds the foreground bounding box of ``study[key]``,
         and uses that bounding box to crop all other images, including segmentations.
 
         Args:
-            key: The key of the image to use for finding the foreground bounding box.
+            key: The key of the image (scan or segmentation) to use for finding the foreground bounding box.
         """
         d = preprocessing.cropping.crop_bg_D(self.get_images(), key)
         return Study(**d, **self.get_info())
 
-    def skullstrip(
+    def skullstrip_hd_bet(
         self,
         key: str,
         register_to_mni152: Literal["T1", "T2"] | None = None,
         device: Literal["cpu", "cuda", "mps"] = CUDA_IF_AVAILABLE,
         disable_tta: bool = False,
         verbose: bool = False,
+
+        expand: int = 0,
+
+        include_mask: bool = False,
+        keep_original: bool = False,
+
     ) -> "Study":
-        """Predicts brain mask of ``study[key]``, then uses this mask to skull strip all scans. Doesn't affect segmentations.
+        """Returns a new study with all scans skullstripped.
+
+        This predicts brain mask of ``study[key]`` via HD-BET,
+        then uses this mask to skullstrip all scans. Doesn't affect segmentations.
 
         Args:
             key: Key of the image to pass to HD-BET for brain mask prediction.
-            register_to_mni152: Should be ``"T1"``, ``"T2"`` or ``None``.
+            register_to_mni152:
+                Modality of MNI152 template to pre-register ``study[key]`` to. Should be ``"T1"``, ``"T2"`` or ``None``.
                 If specified, ``input`` will be registered to specified MNI152 template,
-                and brain mask registered back to original ``input``.
+                then after prediction the brain mask registered back to original ``input``.
                 Note that HD-BET expects images to be in MNI152 space. Defaults to None.
-            device: Used to set on which device the prediction will run. Can be 'cuda' (=GPU), 'cpu' or 'mps'.
+            device:
+                Used to set on which device the prediction will run. Can be 'cuda' (=GPU), 'cpu' or 'mps'.
                 Defaults to CUDA_IF_AVAILABLE.
-            disable_tta: Set this flag to disable test time augmentation. This will make prediction faster
+            disable_tta:
+                Set this flag to disable test time augmentation. This will make prediction faster
                 at a slight decrease in prediction quality. Recommended for device cpu. Defaults to False.
-            verbose: Enable verbose output during processing. Defaults to False.
+            verbose (bool, optional): Talk to me. Defaults to False.
+            expand (int, optional):
+                Positive values expand brain mask by this many pixels, meaning inner parts of the skull will be included;
+                Negative values dilate brain mask by this many pixels, meaning outer parts of the brain will be excluded.
+            include_mask (bool, optional):
+                if True, adds ``"seg_hd_bet"`` with brain mask predicted by HD-BET to returned dictionary.
+                This adds brain mask BEFORE expanding/dilating if ``expand`` argument is specified.
+            keep_original (bool, Optional):
+                if True, skull-stripped images are added to the returned study with ``"_hd_bet"`` postfix,
+                and do not replace original images.
         """
-        d = preprocessing.skullstripping.skullstrip_D(
+        d = preprocessing.hd_bet.skullstrip_D(
             images=self.get_scans(),
             key=key,
             register_to_mni152=register_to_mni152,
             device=device,
             disable_tta=disable_tta,
             verbose=verbose,
+            include_mask=include_mask,
+            keep_original=keep_original,
+            expand=expand,
+        )
+        return Study(**d, **self.get_segmentations(), **self.get_info())
+
+    def skullstrip_synthstrip(
+        self,
+        synthstrip_script_path: str | os.PathLike,
+        key: str,
+        gpu: bool | None = None,
+        border: int | None = None,
+        threads: int | None = None,
+        model: str | os.PathLike | None = None,
+        expand: int = 0,
+
+        include_mask: bool = False,
+        keep_original: bool = False,
+
+        verbose: bool = True,
+    ):
+        """Returns a new study with all scans skullstripped.
+
+        This predicts brain mask of ``study[key]`` via synthstrip,
+        then uses this mask to skullstrip all scans. Doesn't affect segmentations.
+
+        Args:
+            synthstrip_script_path (str | os.PathLike): path to synthstrip script.
+            key (str): key of the image to pass to HD-BET for brain mask prediction.
+            gpu (bool | None, optional): use the GPU, defaults to False if unset.
+            border (int | None, optional): mask border threshold in mm, defaults to 1 if unset.
+            threads (int | None, optional): PyTorch CPU threads, PyTorch default if unset.
+            model (str | os.PathLike | None, optional): alternative model weights
+            expand (int, optional):
+                Positive values expand brain mask by this many pixels, meaning inner parts of the skull will be included;
+                Negative values dilate brain mask by this many pixels, meaning outer parts of the brain will be excluded.
+            include_mask (bool, optional):
+                if True, adds ``"seg_synthstrip"`` with brain mask predicted by HD-BET to returned dictionary.
+                This adds brain mask BEFORE expanding/dilating if ``expand`` argument is specified.
+            keep_original (bool, Optional):
+                if True, skull-stripped images are added to the dictionary
+                with ``"_synthstrip" ``postfix, rather than replacing.
+        """
+        d = preprocessing.synthstrip.skullstrip_D(
+            synthstrip_script_path=synthstrip_script_path,
+            images=self.get_scans(),
+            key=key,
+            gpu=gpu, border=border, threads=threads, model=model,
+            expand=expand, include_mask=include_mask, keep_original=keep_original,
+            verbose=verbose,
         )
         return Study(**d, **self.get_segmentations(), **self.get_info())
 
     def resize(self, size: Sequence[int], interpolator=sitk.sitkLinear):
-        """Resize all images to ``size``.
+        """Returns a new study with all images resized to to ``size``.
 
         Args:
             size: Target size as a sequence of integers (e.g., [height, width, depth]).
-            interpolator: Interpolation method for regular images (segmentations always use nearest neighbor).
+            interpolator: Interpolation method for scans (segmentations always use nearest neighbor).
         """
         return self.apply(
-            partial(preprocessing.registration.resize, new_size=size, interpolator=interpolator,),
-            partial(preprocessing.registration.resize, new_size=size, interpolator=sitk.sitkNearestNeighbor,),
+            partial(preprocessing.resize, new_size=size, interpolator=interpolator,),
+            partial(preprocessing.resize, new_size=size, interpolator=sitk.sitkNearestNeighbor,),
         )
 
     def downsample(self, factor: float, dims = None, interpolator=sitk.sitkLinear):
-        """Downsample all images. Factor = 2 for 2x downsampling. Dims ``None`` for all dims.
+        """Returns a new study with all images downsampled by ``factor`` along ``dims``.
+        For example, ``factor=2`` for 2x downsampling.
+        Set ``dims`` to ``None`` to downsample along all dimensions.
 
         Args:
             factor: Downsampling factor (e.g., 2 for 2x downsampling).
             dims: Specific dimensions to downsample, or None for all dimensions.
-            interpolator: Interpolation method for regular images (segmentations always use nearest neighbor).
+            interpolator: Interpolation method for scans (segmentations always use nearest neighbor).
         """
         return self.apply(
-            partial(preprocessing.registration.downsample, factor=factor, dims=dims, interpolator=interpolator,),
-            partial(preprocessing.registration.downsample, factor=factor, dims=dims, interpolator=sitk.sitkNearestNeighbor,),
+            partial(preprocessing.downsample, factor=factor, dims=dims, interpolator=interpolator,),
+            partial(preprocessing.downsample, factor=factor, dims=dims, interpolator=sitk.sitkNearestNeighbor,),
         )
 
-    def register(self, key: str, to: ImageLike, pmap=None, log_to_console=False) -> "Study":
-        """Returns a new Study, registers ``study[key]`` to ``to``,
+    def register_SE(self, key: str, to: ImageLike, pmap=None, log_to_console=False) -> "Study":
+        """Returns a new Study.
+        Registers ``study[key]`` to ``to`` via SimpleElastix,
         and use transformation parameters to register all other images including segmentation.
-        This assumes that all images are in the same space, if they are not, use ``register_many`` method.
+        This assumes that all images are aligned, if they are not, use ``register_many`` method.
 
         Args:
             key: The key of the image to use as reference for registration.
@@ -213,7 +318,7 @@ class Study(UserDict[str, sitk.Image | Any]):
             pmap: Parameter map for registration. If None, uses default parameters.
             log_to_console: Whether to log registration progress to console.
         """
-        d = preprocessing.registration.register_D(
+        d = preprocessing.simple_elastix.register_D(
             images=self.get_images(),
             key=key,
             to=to,
@@ -222,9 +327,11 @@ class Study(UserDict[str, sitk.Image | Any]):
         )
         return Study(**d, **self.get_info())
 
-    def register_each(self, key: str, to: "ImageLike | None" = None, pmap=None, log_to_console=False) -> "Study":
-        """Returns a new study. Registers all other images to ``study[key]``.
+    def register_each_SE(self, key: str, to: "ImageLike | None" = None, pmap=None, log_to_console=False) -> "Study":
+        """Returns a new study.
+        Registers all other images to ``study[key]``.
         If ``to`` is specified, register ``study[key]`` to ``to`` beforehand.
+        Uses SimpleElastix.
 
         Args:
             key: The key of the image to use as reference for registration.
@@ -242,38 +349,133 @@ class Study(UserDict[str, sitk.Image | Any]):
 
         d = self.get_scans()
         if to is not None:
-            d[key] = preprocessing.registration.register(d[key], to=to, pmap=pmap, log_to_console=log_to_console)
+            d[key] = preprocessing.simple_elastix.register(d[key], to=to, pmap=pmap, log_to_console=log_to_console)
 
         for k in d:
             if k != key:
-                d[k] = preprocessing.registration.register(d[k], to=d[key], pmap=pmap, log_to_console=log_to_console)
+                d[k] = preprocessing.simple_elastix.register(d[k], to=d[key], pmap=pmap, log_to_console=log_to_console)
 
         return Study(**d, **self.get_info())
 
     def resample_to(self, to: "np.ndarray | sitk.Image | torch.Tensor | str", interpolation=sitk.sitkLinear) -> "Study":
-        """Returns a new study, resamples all images including segmentation to `to`.
+        """Returns a new study with all images including segmentation resampled to ``to``.
         Segmentation always uses nearest interpolation"""
         to = tositk(to)
 
         return self.apply(
-            partial(preprocessing.registration.resample_to, to=to, interpolation=interpolation),
-            partial(preprocessing.registration.resample_to, to=to, interpolation=sitk.sitkNearestNeighbor),
+            partial(preprocessing.resample_to, to=to, interpolation=interpolation),
+            partial(preprocessing.resample_to, to=to, interpolation=sitk.sitkNearestNeighbor),
         )
 
-    def n4_bias_field_correction(self, key: str, shrink: int = 4) -> "Study":
+    def n4_bias_field_correction(self, key: str, shrink: int = 4, postfix: str = "") -> "Study":
         """Returns a new study with corrected bias field of the image under ``key``. Doesn't affect other images.
 
         Args:
-            key: The key of the image for which to correct the bias field.
-            shrink: By how many times to shrink the size of input image for calculating the bias field.
+            key (str): The key of the image for which to correct the bias field.
+            shrink (int, optional): By how many times to shrink the size of input image for calculating the bias field.
                 The bias field is then applied to original size (unshrunk) image.
                 Setting shrink to 1 disables it, but n4 algorithm may take several minutes.
                 Setting it to ~4 is good enough in most cases and will be significantly faster (usually few seconds).
+            postfix (str, optional):
+                if specified, N4-corrected image is added to returned study with specified postfix rather than
+                replacing current ``key``.
         """
         new = self.copy()
-        new[key] = preprocessing.bias_field_correction.n4_bias_field_correction(new[key], shrink=shrink)
+        new[f"{key}{postfix}"] = preprocessing.bias_field_correction.n4_bias_field_correction(new[key], shrink=shrink)
         return new
 
+    def expand_binary_mask(self, key: str, expand: int, postfix: str = ""):
+        """Returns a new study with binary mask under ``key`` expanded or dilated by ``expand`` pixels.
+
+        Args:
+            key (str): The key of the mask to expand/dilate.
+            expand (int, optional):
+                Positive values expand the mask by this many pixels;
+                Negative values dilate the mask by this many pixels.
+            postfix (str, optional):
+                if specified, expanded/dilated mask is added to returned study with specified postfix rather than
+                replacing current ``key``.
+        """
+        new = self.copy()
+        new[f"{key}{postfix}"] = preprocessing.mask.expand_binary_mask(new[key], expand=expand)
+        return new
+
+    def remove_small_objects(
+        self,
+        key: str,
+        min_size=64,
+        connectivity=1,
+        independent_channels: bool = False,
+        postfix: str = "",
+    ):
+        """Returns a new study with small objects removed in mask under ``key``,
+        this uses ``skimage.morphology.remove-small-objects`` or
+        ``monai.transforms.remove_small_objects``.
+
+        See: https://scikit-image.org/docs/dev/api/skimage.morphology.html#remove-small-objects.
+
+        Args:
+            key (str): The key of the mask to remove small objects in.
+            min_size (int, optional): objects smaller than this size are removed. Defaults to 64.
+            connectivity (int, optional):
+                Maximum number of orthogonal hops to consider a pixel/voxel as a neighbor.
+                Accepted values are ranging from 1 to input.ndim.
+                If None, a full connectivity of input.ndim is used.
+                For more details refer to linked scikit-image documentation. Defaults to 1.
+            independent_channels (bool, optional):
+                Whether to consider each channel independently (requires monai). Defaults to False.
+            postfix (str, optional):
+                if specified, processed mask is added to returned study with specified postfix rather than
+                replacing current ``key``.
+        """
+        arr = self.numpy(key)
+
+        if independent_channels:
+            from monai.transforms import remove_small_objects
+            # Data should be one-hotted.
+            num_classes = arr.max() + 1
+            one_hot = np.zeros((num_classes, *arr.shape))
+            i, j, k = np.indices(arr.shape)
+            one_hot[arr[i,j,k], i, j, k] = 1
+            one_hot = remove_small_objects(
+                one_hot, min_size=min_size, connectivity=connectivity, independent_channels=independent_channels)
+            arr = one_hot.argmax(0)
+            return self.add(f'{key}{postfix}', arr, reference_key=key)
+
+        from skimage.morphology import remove_small_objects
+        binary = (arr != 0)
+        arr *= remove_small_objects(binary, min_size=min_size, connectivity=connectivity)
+        return self.add(f'{key}{postfix}', arr, reference_key=key)
+
+    def keep_largest_connected_component(self, key: str, applied_labels=None, independent=True, connectivity=None, num_components=1, postfix: str = ""):
+        """Returns a new study with largest components kept in mask under ``key``,
+        this uses ``monai.transforms.KeepLargestConnectedComponent``.
+
+        Args:
+            key (str): The key of the mask to remove keep largest components in.
+            applied_labels: Labels for applying the connected component analysis on.
+                If given, voxels whose value is in this list will be analyzed.
+                If `None`, all non-zero values will be analyzed.
+            independent: whether to treat ``applied_labels`` as a union of foreground labels.
+                If ``True``, the connected component analysis will be performed on each foreground label independently
+                and return the intersection of the largest components.
+                If ``False``, the analysis will be performed on the union of foreground labels.
+                default is `True`.
+            connectivity: Maximum number of orthogonal hops to consider a pixel/voxel as a neighbor.
+                Accepted values are ranging from  1 to input.ndim. If ``None``, a full
+                connectivity of ``input.ndim`` is used. for more details:
+                https://scikit-image.org/docs/dev/api/skimage.measure.html#skimage.measure.label.
+            num_components: The number of largest components to preserve.
+            postfix (str, optional):
+                if specified, processed mask is added to returned study with specified postfix rather than
+                replacing current ``key``.
+        """
+        from monai.transforms import KeepLargestConnectedComponent
+        tfm = KeepLargestConnectedComponent(applied_labels=applied_labels, is_onehot=False, independent=independent,
+                                            connectivity=connectivity, num_components=num_components)
+        arr = tfm(self.numpy(key))
+        return self.add(f'{key}{postfix}', arr, reference_key=key)
+
     def numpy(self, key: str):
         """returns ``study[key]`` converted to a numpy array."""
         return tonumpy(self[key])
@@ -339,8 +541,8 @@ class Study(UserDict[str, sitk.Image | Any]):
         return {k: (torch.from_numpy(v) if isinstance(v, np.ndarray) else v) for k,v in self.numpy_dict()}
 
     def plot(self):
-        from .utils.plotting import visualize_3d_arrays
-        visualize_3d_arrays(self.get_images().numpy_dict())
+        from .utils.plotting import plot_study
+        return plot_study(self.get_images().numpy_dict())
 
     def save(
         self,
